import time
from urllib.parse import parse_qs, urlparse

from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import SuspiciousOperation
from django.test import Client, RequestFactory, TestCase, override_settings
from django.urls import reverse
from unittest.mock import patch

from mozilla_django_oidc import views

TEST_CODE_VERIFIER = "ThisStringIsURLSafeAndAtLeast43Characters00"


User = get_user_model()


def my_custom_op_logout(request):
    return request.build_absolute_uri("/logged/out")


class OIDCAuthorizationCallbackViewTestCase(TestCase):
    def setUp(self):
        self.factory = RequestFactory()

    @override_settings(LOGIN_REDIRECT_URL="/success")
    def test_get_auth_success(self):
        """Test successful callback request to RP."""
        user = User.objects.create_user("example_username")

        get_data = {"code": "example_code", "state": "example_state"}
        url = reverse("oidc_authentication_callback")
        request = self.factory.get(url, get_data)
        client = Client()
        request.session = client.session
        request.session["oidc_states"] = {
            "example_state": {
                "code_verifier": TEST_CODE_VERIFIER,
                "nonce": None,
                "added_on": time.time(),
            },
        }
        callback_view = views.OIDCAuthenticationCallbackView.as_view()

        with patch("mozilla_django_oidc.views.auth.authenticate") as mock_auth:
            with patch("mozilla_django_oidc.views.auth.login") as mock_login:
                mock_auth.return_value = user
                response = callback_view(request)

                mock_auth.assert_called_once_with(
                    code_verifier=TEST_CODE_VERIFIER, nonce=None, request=request
                )
                mock_login.assert_called_once_with(request, user)

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/success")

    @override_settings(LOGIN_REDIRECT_URL="/success")
    def test_get_auth_success_next_url(self):
        """Test successful callback request to RP with custom `next` url."""
        user = User.objects.create_user("example_username")

        get_data = {"code": "example_code", "state": "example_state"}
        url = reverse("oidc_authentication_callback")
        request = self.factory.get(url, get_data)
        client = Client()
        request.session = client.session
        request.session["oidc_states"] = {
            "example_state": {
                "code_verifier": TEST_CODE_VERIFIER,
                "nonce": None,
                "added_on": time.time(),
            },
        }
        request.session["oidc_login_next"] = "/foobar"
        callback_view = views.OIDCAuthenticationCallbackView.as_view()

        with patch("mozilla_django_oidc.views.auth.authenticate") as mock_auth:
            with patch("mozilla_django_oidc.views.auth.login") as mock_login:
                mock_auth.return_value = user
                response = callback_view(request)

                mock_auth.assert_called_once_with(
                    code_verifier=TEST_CODE_VERIFIER, nonce=None, request=request
                )
                mock_login.assert_called_once_with(request, user)

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/foobar")

    @override_settings(LOGIN_REDIRECT_URL="/success")
    def test_get_auth_success_without_pkce(self):
        """Test successful callback request to RP after disabling PKCE."""
        user = User.objects.create_user("example_username")

        get_data = {"code": "example_code", "state": "example_state"}
        url = reverse("oidc_authentication_callback")
        request = self.factory.get(url, get_data)
        client = Client()
        request.session = client.session
        request.session["oidc_states"] = {
            "example_state": {
                # Simulates an auth request sent with PKCE disabled
                # 'code_verifier': TEST_CODE_VERIFIER,
                "nonce": None,
                "added_on": time.time(),
            },
        }
        callback_view = views.OIDCAuthenticationCallbackView.as_view()

        with patch("mozilla_django_oidc.views.auth.authenticate") as mock_auth:
            with patch("mozilla_django_oidc.views.auth.login") as mock_login:
                mock_auth.return_value = user
                response = callback_view(request)

                mock_auth.assert_called_once_with(
                    code_verifier=None, nonce=None, request=request
                )
                mock_login.assert_called_once_with(request, user)

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/success")

    @override_settings(LOGIN_REDIRECT_URL_FAILURE="/failure")
    def test_get_auth_failure_nonexisting_user(self):
        """Test unsuccessful authentication and redirect url."""
        get_data = {"code": "example_code", "state": "example_state"}

        url = reverse("oidc_authentication_callback")
        request = self.factory.get(url, get_data)
        client = Client()
        request.session = client.session
        request.session["oidc_states"] = {
            "example_state": {"nonce": None, "added_on": time.time()},
        }
        callback_view = views.OIDCAuthenticationCallbackView.as_view()

        with patch("mozilla_django_oidc.views.auth.authenticate") as mock_auth:
            mock_auth.return_value = None
            response = callback_view(request)

            mock_auth.assert_called_once_with(
                code_verifier=None, nonce=None, request=request
            )

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/failure")

    @override_settings(LOGIN_REDIRECT_URL_FAILURE="/failure")
    def test_get_auth_failure_inactive_user(self):
        """Test authentication failure attempt for an inactive user."""
        user = User.objects.create_user("example_username")
        user.is_active = False
        user.save()

        get_data = {"code": "example_code", "state": "example_state"}

        url = reverse("oidc_authentication_callback")
        request = self.factory.get(url, get_data)
        client = Client()
        request.session = client.session
        request.session["oidc_states"] = {
            "example_state": {"nonce": None, "added_on": time.time()},
        }
        callback_view = views.OIDCAuthenticationCallbackView.as_view()

        with patch("mozilla_django_oidc.views.auth.authenticate") as mock_auth:
            mock_auth.return_value = user
            response = callback_view(request)

            mock_auth.assert_called_once_with(
                code_verifier=None, request=request, nonce=None
            )

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/failure")

    @override_settings(LOGIN_REDIRECT_URL_FAILURE="/failure")
    def test_get_auth_error(self):
        """Test authentication error handling.

        Sttate should be removed from session and user should be logged out.
        """
        user = User.objects.create_user("example_username")

        get_data = {"error": "example_code", "state": "example_state"}
        url = reverse("oidc_authentication_callback")
        request = self.factory.get(url, get_data)
        client = Client()
        request.session = client.session
        request.session["oidc_states"] = {
            "example_state": {"nonce": None, "added_on": time.time()},
        }
        request.user = user
        callback_view = views.OIDCAuthenticationCallbackView.as_view()

        with patch("mozilla_django_oidc.views.auth.logout") as mock_logout:

            def clear_user(request):
                # Assert state is cleared prior to logout
                self.assertEqual(request.session["oidc_states"], {})
                request.user = AnonymousUser()

            mock_logout.side_effect = clear_user
            response = callback_view(request)
            mock_logout.assert_called_once()

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/failure")

    @override_settings(OIDC_USE_NONCE=False)
    @override_settings(LOGIN_REDIRECT_URL_FAILURE="/failure")
    def test_get_auth_dirty_data(self):
        """Test authentication attempt with wrong get data."""
        get_data = {
            "foo": "bar",
        }

        url = reverse("oidc_authentication_callback")
        request = self.factory.get(url, get_data)
        request.session = {}
        callback_view = views.OIDCAuthenticationCallbackView.as_view()
        response = callback_view(request)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/failure")

    @override_settings(LOGIN_REDIRECT_URL_FAILURE="/failure")
    def test_get_auth_failure_missing_session_state(self):
        """Test authentication failure attempt for an inactive user."""
        user = User.objects.create_user("example_username")
        user.is_active = False
        user.save()

        get_data = {"code": "example_code", "state": "example_state"}

        url = reverse("oidc_authentication_callback")
        request = self.factory.get(url, get_data)
        request.session = {}
        callback_view = views.OIDCAuthenticationCallbackView.as_view()

        response = callback_view(request)

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/failure")

    @override_settings(LOGIN_REDIRECT_URL_FAILURE="/failure")
    def test_get_auth_failure_tampered_session_state(self):
        """Test authentication failure attempt for an inactive user."""
        user = User.objects.create_user("example_username")
        user.is_active = False
        user.save()

        get_data = {"code": "example_code", "state": "example_state"}

        url = reverse("oidc_authentication_callback")
        request = self.factory.get(url, get_data)
        request.session = {"oidc_states": {"tampered_state": None}}
        callback_view = views.OIDCAuthenticationCallbackView.as_view()

        with self.assertRaises(SuspiciousOperation) as context:
            callback_view(request)

        expected_error_message = (
            "OIDC callback state not found in session `oidc_states`!"
        )
        self.assertEqual(context.exception.args, (expected_error_message,))

    @override_settings(LOGIN_REDIRECT_URL="/success")
    def test_nonce_is_deleted(self):
        """Test Nonce is not in session."""
        user = User.objects.create_user("example_username")

        get_data = {"code": "example_code", "state": "example_state"}
        url = reverse("oidc_authentication_callback")
        request = self.factory.get(url, get_data)
        client = Client()
        request.session = client.session
        request.session["oidc_states"] = {
            "example_state": {"nonce": "example_nonce", "added_on": time.time()},
        }
        request.session["oidc_nonce"] = "example_nonce"
        callback_view = views.OIDCAuthenticationCallbackView.as_view()

        with patch("mozilla_django_oidc.views.auth.authenticate") as mock_auth:
            with patch("mozilla_django_oidc.views.auth.login") as mock_login:
                mock_auth.return_value = user
                response = callback_view(request)

                mock_auth.assert_called_once_with(
                    code_verifier=None, nonce="example_nonce", request=request
                )
                mock_login.assert_called_once_with(request, user)

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/success")
        self.assertTrue("example_state" not in request.session["oidc_states"])

    def test_multiple_login_sessions(self):
        """Test if states/nonces of other login sessions remain in the 'oidc_states' dictionary."""
        get_data = {"code": "example_code", "state": "example_state"}
        url = reverse("oidc_authentication_callback")
        request = self.factory.get(url, get_data)
        client = Client()
        request.session = client.session
        request.session["oidc_states"] = {
            "example_state": {"nonce": None, "added_on": time.time()},
            "example_state2": {"nonce": None, "added_on": time.time()},
        }
        callback_view = views.OIDCAuthenticationCallbackView.as_view()
        callback_view(request)

        self.assertFalse("example_state" in request.session["oidc_states"])
        self.assertTrue("example_state2" in request.session["oidc_states"])

    @override_settings(LOGIN_REDIRECT_URL="/success")
    def test_session_refresh_doesnt_call_login_again(self):
        """Test that a successful session refresh doesn't call django.contrib.auth.login
        again (which would cycle session keys and CSRF tokens)."""
        user = User.objects.create_user("example_username")

        def make_request():
            get_data = {"code": "example_code", "state": "example_state"}
            url = reverse("oidc_authentication_callback")
            request = self.factory.get(url, get_data)
            client = Client()
            request.session = client.session
            request.session["oidc_states"] = {
                "example_state": {"nonce": None, "added_on": time.time()},
            }
            return request

        request = make_request()
        callback_view = views.OIDCAuthenticationCallbackView.as_view()

        # Login the first time. This should call login()
        with patch("mozilla_django_oidc.views.auth.authenticate") as mock_auth:
            with patch("mozilla_django_oidc.views.auth.login") as mock_login:
                mock_auth.return_value = user
                response = callback_view(request)
                mock_auth.assert_called_once_with(
                    code_verifier=None, nonce=None, request=request
                )
                mock_login.assert_called_once_with(request, user)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/success")

        # Set the user on the request, to mimic a session refresh
        request = make_request()
        request.user = user

        # Login the second time. This should succeed, but not call login(), since the
        # user is already logged in.
        with patch("mozilla_django_oidc.views.auth.authenticate") as mock_auth:
            with patch("mozilla_django_oidc.views.auth.login") as mock_login:
                mock_auth.return_value = user
                response = callback_view(request)
                mock_auth.assert_called_once_with(
                    code_verifier=None, nonce=None, request=request
                )
                mock_login.assert_not_called()
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/success")


class GetNextURLTestCase(TestCase):
    def setUp(self):
        self.factory = RequestFactory()

    def build_request(self, next_url):
        return self.factory.get("/", data={"next": next_url})

    def test_no_param(self):
        req = self.factory.get("/")
        next_url = views.get_next_url(req, "next")
        self.assertIsNone(next_url)

    def test_non_next_param(self):
        req = self.factory.get("/", data={"redirectto": "/foo"})
        next_url = views.get_next_url(req, "redirectto")
        self.assertEqual(next_url, "/foo")

    def test_good_urls(self):
        urls = [
            "/",
            "/foo",
            "/foo?bar=baz",
            "http://testserver/foo",
        ]
        for url in urls:
            req = self.build_request(next_url=url)
            next_url = views.get_next_url(req, "next")

            self.assertEqual(next_url, url)

    def test_bad_urls(self):
        urls = [
            "",
            # NOTE(willkg): Test data taken from the Django url_has_allowed_host_and_scheme tests.
            "http://example.com",
            "http:///example.com",
            "https://example.com",
            "ftp://example.com",
            r"\\example.com",
            r"\\\example.com",
            r"/\\/example.com",
            r"\\\example.com",
            r"\\example.com",
            r"\\//example.com",
            r"/\/example.com",
            r"\/example.com",
            r"/\example.com",
            "http:///example.com",
            r"http:/\//example.com",
            r"http:\/example.com",
            r"http:/\example.com",
            'javascript:alert("XSS")',
            "\njavascript:alert(x)",
            "\x08//example.com",
            r"http://otherserver\@example.com",
            r"http:\\testserver\@example.com",
            r"http://testserver\me:pass@example.com",
            r"http://testserver\@example.com",
            r"http:\\testserver\confirm\me@example.com",
            "http:999999999",
            "ftp:9999999999",
            "\n",
        ]
        for url in urls:
            req = self.build_request(next_url=url)
            next_url = views.get_next_url(req, "next")

            self.assertIsNone(next_url)

    def test_https(self):
        # If the request is for HTTPS and the next url is HTTPS, then that
        # works with all Djangos.
        req = self.factory.get(
            "/",
            data={"next": "https://testserver/foo"},
            secure=True,
        )
        self.assertEqual(req.is_secure(), True)
        next_url = views.get_next_url(req, "next")
        self.assertEqual(next_url, "https://testserver/foo")

        # If the request is for HTTPS and the next url is HTTP, then that fails.
        req = self.factory.get(
            "/",
            data={"next": "http://testserver/foo"},
            secure=True,
        )
        self.assertEqual(req.is_secure(), True)
        next_url = views.get_next_url(req, "next")
        self.assertIsNone(next_url)

    @override_settings(OIDC_REDIRECT_REQUIRE_HTTPS=False)
    def test_redirect_https_not_required(self):
        req = self.factory.get("/", data={"next": "http://testserver/foo"}, secure=True)

        next_url = views.get_next_url(req, "next")
        self.assertEqual(next_url, "http://testserver/foo")

    @override_settings(OIDC_REDIRECT_ALLOWED_HOSTS=["example.com", "foo.com"])
    def test_redirect_allowed_hosts(self):
        req = self.factory.get(
            "/", data={"next": "https://example.com/foo"}, secure=True
        )

        next_url = views.get_next_url(req, "next")
        self.assertEqual(next_url, "https://example.com/foo")


class OIDCAuthorizationRequestViewTestCase(TestCase):
    def setUp(self):
        self.factory = RequestFactory()

    @override_settings(OIDC_OP_AUTHORIZATION_ENDPOINT="https://server.example.com/auth")
    @override_settings(OIDC_RP_CLIENT_ID="example_id")
    @override_settings(OIDC_USE_PKCE=True)
    @patch("mozilla_django_oidc.views.get_random_string")
    def test_get(self, mock_views_random):
        """Test initiation of a successful OIDC attempt."""
        mock_views_random.return_value = "examplestring"
        url = reverse("oidc_authentication_init")
        request = self.factory.get(url)
        request.session = dict()
        login_view = views.OIDCAuthenticationRequestView.as_view()
        response = login_view(request)
        self.assertEqual(response.status_code, 302)

        o = urlparse(response.url)
        query_dict = parse_qs(o.query)

        # The PKCE code_challenge should be a random string between 43 and 128 characters.
        # Since it's random, we can only test that it's present and has the right length.
        # Then we just insert it into the expected_query.
        self.assertIn("code_challenge", query_dict)
        self.assertTrue(
            len(query_dict["code_challenge"]) == 1
            and 43 <= len(query_dict["code_challenge"][0]) <= 128
        )
        expected_query = {
            "code_challenge": query_dict["code_challenge"],
            "code_challenge_method": ["S256"],
            "response_type": ["code"],
            "scope": ["openid email"],
            "client_id": ["example_id"],
            "redirect_uri": ["http://testserver/callback/"],
            "state": ["examplestring"],
            "nonce": ["examplestring"],
        }
        self.assertDictEqual(query_dict, expected_query)
        self.assertEqual(o.hostname, "server.example.com")
        self.assertEqual(o.path, "/auth")

    @override_settings(OIDC_OP_AUTHORIZATION_ENDPOINT="https://server.example.com/auth")
    @override_settings(OIDC_RP_CLIENT_ID="example_id")
    @override_settings(OIDC_USE_PKCE=False)
    @patch("mozilla_django_oidc.views.get_random_string")
    def test_get_without_PKCE(self, mock_views_random):
        """Test initiation of a successful OIDC attempt with PKCE disabled."""
        mock_views_random.return_value = "examplestring"
        url = reverse("oidc_authentication_init")
        request = self.factory.get(url)
        request.session = dict()
        login_view = views.OIDCAuthenticationRequestView.as_view()
        response = login_view(request)
        self.assertEqual(response.status_code, 302)

        o = urlparse(response.url)
        query_dict = parse_qs(o.query)

        # PKCE is disabled, so code_challenge and code_challenge_method should not be present.
        expected_query = {
            "response_type": ["code"],
            "scope": ["openid email"],
            "client_id": ["example_id"],
            "redirect_uri": ["http://testserver/callback/"],
            "state": ["examplestring"],
            "nonce": ["examplestring"],
        }
        self.assertDictEqual(query_dict, expected_query)
        self.assertEqual(o.hostname, "server.example.com")
        self.assertEqual(o.path, "/auth")

    @override_settings(OIDC_OP_AUTHORIZATION_ENDPOINT="https://server.example.com/auth")
    @override_settings(OIDC_RP_CLIENT_ID="example_id")
    @override_settings(OIDC_USE_PKCE=True)
    @override_settings(OIDC_PKCE_CODE_VERIFIER_SIZE=42)  # must be between 43 and 128
    @patch("mozilla_django_oidc.views.get_random_string")
    def test_get_invalid_code_verifier_size_too_short(self, mock_views_random):
        """Test initiation of an OIDC attempt with an invalid code verifier size."""
        mock_views_random.return_value = "examplestring"
        url = reverse("oidc_authentication_init")
        request = self.factory.get(url)
        request.session = dict()
        login_view = views.OIDCAuthenticationRequestView.as_view()
        try:
            login_view(request)
            self.fail(
                "OIDC_PKCE_CODE_VERIFIER_SIZE must be between 43 and 128,"
                " but OIDC_PKCE_CODE_VERIFIER_SIZE was 42 and no exception was raised."
            )
        except ValueError:
            pass

    @override_settings(OIDC_OP_AUTHORIZATION_ENDPOINT="https://server.example.com/auth")
    @override_settings(OIDC_RP_CLIENT_ID="example_id")
    @override_settings(OIDC_USE_PKCE=True)
    @override_settings(OIDC_PKCE_CODE_VERIFIER_SIZE=129)  # must be between 43 and 128
    @patch("mozilla_django_oidc.views.get_random_string")
    def test_get_invalid_code_verifier_size_too_long(self, mock_views_random):
        """Test initiation of an OIDC attempt with an invalid code verifier size."""
        mock_views_random.return_value = "examplestring"
        url = reverse("oidc_authentication_init")
        request = self.factory.get(url)
        request.session = dict()
        login_view = views.OIDCAuthenticationRequestView.as_view()
        try:
            login_view(request)
            self.fail(
                "OIDC_PKCE_CODE_VERIFIER_SIZE must be between 43 and 128,"
                " but OIDC_PKCE_CODE_VERIFIER_SIZE was 129 and no exception was raised."
            )
        except ValueError:
            pass

    @override_settings(ROOT_URLCONF="tests.namespaced_urls")
    @override_settings(OIDC_OP_AUTHORIZATION_ENDPOINT="https://server.example.com/auth")
    @override_settings(OIDC_RP_CLIENT_ID="example_id")
    @override_settings(OIDC_USE_PKCE=True)
    @override_settings(
        OIDC_AUTHENTICATION_CALLBACK_URL="namespace:oidc_authentication_callback"
    )
    @patch("mozilla_django_oidc.views.get_random_string")
    def test_get_namespaced(self, mock_views_random):
        """Test initiation of a successful OIDC attempt with namespaced redirect_uri."""
        mock_views_random.return_value = "examplestring"
        url = reverse("namespace:oidc_authentication_init")
        request = self.factory.get(url)
        request.session = dict()
        login_view = views.OIDCAuthenticationRequestView.as_view()
        response = login_view(request)
        self.assertEqual(response.status_code, 302)

        o = urlparse(response.url)
        query_dict = parse_qs(o.query)

        # The PKCE code_challenge should be a random string between 43 and 128 characters.
        # Since it's random, we can only test that it's present and has the right length.
        # Then we just insert it into the expected_query.
        self.assertIn("code_challenge", query_dict)
        self.assertTrue(
            len(query_dict["code_challenge"]) == 1
            and 43 <= len(query_dict["code_challenge"][0]) <= 128
        )
        expected_query = {
            "code_challenge": query_dict["code_challenge"],
            "code_challenge_method": ["S256"],
            "response_type": ["code"],
            "scope": ["openid email"],
            "client_id": ["example_id"],
            "redirect_uri": ["http://testserver/namespace/callback/"],
            "state": ["examplestring"],
            "nonce": ["examplestring"],
        }
        self.assertDictEqual(query_dict, expected_query)
        self.assertEqual(o.hostname, "server.example.com")
        self.assertEqual(o.path, "/auth")

    @override_settings(OIDC_OP_AUTHORIZATION_ENDPOINT="https://server.example.com/auth")
    @override_settings(OIDC_RP_CLIENT_ID="example_id")
    @override_settings(OIDC_USE_PKCE=True)
    @override_settings(
        OIDC_AUTH_REQUEST_EXTRA_PARAMS={"audience": "some-api.example.com"}
    )
    @patch("mozilla_django_oidc.views.get_random_string")
    def test_get_with_audience(self, mock_views_random):
        """Test initiation of a successful OIDC attempt."""
        mock_views_random.return_value = "examplestring"
        url = reverse("oidc_authentication_init")
        request = self.factory.get(url)
        request.session = dict()
        login_view = views.OIDCAuthenticationRequestView.as_view()
        response = login_view(request)
        self.assertEqual(response.status_code, 302)

        o = urlparse(response.url)
        query_dict = parse_qs(o.query)

        # The PKCE code_challenge should be a random string between 43 and 128 characters.
        # Since it's random, we can only test that it's present and has the right length.
        # Then we just insert it into the expected_query.
        self.assertIn("code_challenge", query_dict)
        self.assertTrue(
            len(query_dict["code_challenge"]) == 1
            and 43 <= len(query_dict["code_challenge"][0]) <= 128
        )
        expected_query = {
            "code_challenge": query_dict["code_challenge"],
            "code_challenge_method": ["S256"],
            "response_type": ["code"],
            "scope": ["openid email"],
            "client_id": ["example_id"],
            "redirect_uri": ["http://testserver/callback/"],
            "state": ["examplestring"],
            "nonce": ["examplestring"],
            "audience": ["some-api.example.com"],
        }
        self.assertDictEqual(query_dict, expected_query)
        self.assertEqual(o.hostname, "server.example.com")
        self.assertEqual(o.path, "/auth")

    @override_settings(OIDC_OP_AUTHORIZATION_ENDPOINT="https://server.example.com/auth")
    @override_settings(OIDC_RP_CLIENT_ID="example_id")
    @override_settings(OIDC_USE_PKCE=True)
    @patch("mozilla_django_oidc.views.get_random_string")
    @patch("mozilla_django_oidc.views.OIDCAuthenticationRequestView.get_extra_params")
    def test_get_with_overridden_extra_params(
        self, mock_extra_params, mock_views_random
    ):
        """Test overriding OIDCAuthenticationRequestView.get_extra_params()."""
        mock_views_random.return_value = "examplestring"

        mock_extra_params.return_value = {"connection": "foo"}

        url = reverse("oidc_authentication_init")
        request = self.factory.get(url)
        request.session = dict()
        login_view = views.OIDCAuthenticationRequestView.as_view()
        response = login_view(request)
        self.assertEqual(response.status_code, 302)

        o = urlparse(response.url)
        query_dict = parse_qs(o.query)

        # The PKCE code_challenge should be a random string between 43 and 128 characters.
        # Since it's random, we can only test that it's present and has the right length.
        # Then we just insert it into the expected_query.
        self.assertIn("code_challenge", query_dict)
        self.assertTrue(
            len(query_dict["code_challenge"]) == 1
            and 43 <= len(query_dict["code_challenge"][0]) <= 128
        )
        expected_query = {
            "code_challenge": query_dict["code_challenge"],
            "code_challenge_method": ["S256"],
            "response_type": ["code"],
            "scope": ["openid email"],
            "client_id": ["example_id"],
            "redirect_uri": ["http://testserver/callback/"],
            "state": ["examplestring"],
            "nonce": ["examplestring"],
            "connection": ["foo"],
        }
        self.assertDictEqual(parse_qs(o.query), expected_query)
        self.assertEqual(o.hostname, "server.example.com")
        self.assertEqual(o.path, "/auth")

        mock_extra_params.assert_called_with(request)

    @override_settings(OIDC_OP_AUTHORIZATION_ENDPOINT="https://server.example.com/auth")
    @override_settings(OIDC_RP_CLIENT_ID="example_id")
    def test_next_url(self):
        """Test that `next` url gets stored to user session."""
        url = reverse("oidc_authentication_init")
        request = self.factory.get("{url}?{params}".format(url=url, params="next=/foo"))
        request.session = dict()
        login_view = views.OIDCAuthenticationRequestView.as_view()
        login_view(request)
        self.assertTrue("oidc_login_next" in request.session)
        self.assertEqual(request.session["oidc_login_next"], "/foo")

    @override_settings(OIDC_OP_AUTHORIZATION_ENDPOINT="https://server.example.com/auth")
    @override_settings(OIDC_RP_CLIENT_ID="example_id")
    def test_missing_next_url(self):
        """Test that `next` url gets invalidated in user session."""
        url = reverse("oidc_authentication_init")
        request = self.factory.get(url)
        request.session = {"oidc_login_next": "foobar"}
        login_view = views.OIDCAuthenticationRequestView.as_view()
        login_view(request)
        self.assertTrue("oidc_login_next" in request.session)
        self.assertTrue(request.session["oidc_login_next"] is None)


class OIDCLogoutViewTestCase(TestCase):
    def setUp(self):
        self.factory = RequestFactory()

    @override_settings(LOGOUT_REDIRECT_URL="/example-logout")
    def test_get_anonymous_user(self):
        url = reverse("oidc_logout")
        request = self.factory.post(url)
        request.user = AnonymousUser()
        logout_view = views.OIDCLogoutView.as_view()

        response = logout_view(request)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/example-logout")

    @override_settings(LOGOUT_REDIRECT_URL="/example-logout")
    def test_post(self):
        user = User.objects.create_user("example_username")
        url = reverse("oidc_logout")
        request = self.factory.post(url)
        request.user = user
        logout_view = views.OIDCLogoutView.as_view()

        with patch("mozilla_django_oidc.views.auth.logout") as mock_logout:
            response = logout_view(request)
            mock_logout.assert_called_once_with(request)

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/example-logout")

    @override_settings(LOGOUT_REDIRECT_URL="/example-logout")
    @override_settings(OIDC_OP_LOGOUT_URL_METHOD="tests.test_views.my_custom_op_logout")
    def test_post_with_OIDC_OP_LOGOUT_URL_METHOD(self):
        user = User.objects.create_user("example_username")
        url = reverse("oidc_logout")
        request = self.factory.post(url)
        request.user = user
        logout_view = views.OIDCLogoutView.as_view()

        with patch("mozilla_django_oidc.views.auth.logout") as mock_logout:
            response = logout_view(request)
            mock_logout.assert_called_once_with(request)

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "http://testserver/logged/out")



import base64
import hashlib
import json
import logging

import inspect
import requests
from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend
from django.core.exceptions import ImproperlyConfigured, SuspiciousOperation
from django.urls import reverse
from django.utils.encoding import force_bytes, smart_bytes, smart_str
from django.utils.module_loading import import_string
from josepy.b64 import b64decode
from josepy.jwk import JWK
from josepy.jws import JWS, Header
from requests.auth import HTTPBasicAuth
from requests.exceptions import HTTPError

from mozilla_django_oidc.utils import absolutify, import_from_settings

LOGGER = logging.getLogger(__name__)


def default_username_algo(email, claims=None):
    """Generate username for the Django user.

    :arg str/unicode email: the email address to use to generate a username
    :arg dic claims: the claims from your OIDC provider, currently unused

    :returns: str/unicode

    """
    # bluntly stolen from django-browserid
    # store the username as a base64 encoded sha224 of the email address
    # this protects against data leakage because usernames are often
    # treated as public identifiers (so we can't use the email address).
    username = base64.urlsafe_b64encode(
        hashlib.sha1(force_bytes(email)).digest()
    ).rstrip(b"=")

    return smart_str(username)


class OIDCAuthenticationBackend(ModelBackend):
    """Override Django's authentication."""

    def __init__(self, *args, **kwargs):
        """Initialize settings."""
        self.OIDC_OP_TOKEN_ENDPOINT = self.get_settings("OIDC_OP_TOKEN_ENDPOINT")
        self.OIDC_OP_USER_ENDPOINT = self.get_settings("OIDC_OP_USER_ENDPOINT")
        self.OIDC_OP_JWKS_ENDPOINT = self.get_settings("OIDC_OP_JWKS_ENDPOINT", None)
        self.OIDC_RP_CLIENT_ID = self.get_settings("OIDC_RP_CLIENT_ID")
        self.OIDC_RP_CLIENT_SECRET = self.get_settings("OIDC_RP_CLIENT_SECRET")
        self.OIDC_RP_SIGN_ALGO = self.get_settings("OIDC_RP_SIGN_ALGO", "HS256")
        self.OIDC_RP_IDP_SIGN_KEY = self.get_settings("OIDC_RP_IDP_SIGN_KEY", None)

        if (
            self.OIDC_RP_SIGN_ALGO.startswith("RS")
            or self.OIDC_RP_SIGN_ALGO.startswith("ES")
        ) and (
            self.OIDC_RP_IDP_SIGN_KEY is None and self.OIDC_OP_JWKS_ENDPOINT is None
        ):
            msg = "{} alg requires OIDC_RP_IDP_SIGN_KEY or OIDC_OP_JWKS_ENDPOINT to be configured."
            raise ImproperlyConfigured(msg.format(self.OIDC_RP_SIGN_ALGO))

        self.UserModel = get_user_model()

    @staticmethod
    def get_settings(attr, *args):
        return import_from_settings(attr, *args)

    def describe_user_by_claims(self, claims):
        email = claims.get("email")
        return "email {}".format(email)

    def filter_users_by_claims(self, claims):
        """Return all users matching the specified email."""
        email = claims.get("email")
        if not email:
            return self.UserModel.objects.none()
        return self.UserModel.objects.filter(email__iexact=email)

    def verify_claims(self, claims):
        """Verify the provided claims to decide if authentication should be allowed."""

        # Verify claims required by default configuration
        scopes = self.get_settings("OIDC_RP_SCOPES", "openid email")
        if "email" in scopes.split():
            return "email" in claims

        LOGGER.warning(
            "Custom OIDC_RP_SCOPES defined. "
            "You need to override `verify_claims` for custom claims verification."
        )

        return True

    def create_user(self, claims):
        """Return object for a newly created user account."""
        email = claims.get("email")
        username = self.get_username(claims)
        return self.UserModel.objects.create_user(username, email=email)

    def get_username(self, claims):
        """Generate username based on claims."""
        # bluntly stolen from django-browserid
        # https://github.com/mozilla/django-browserid/blob/master/django_browserid/auth.py

        username_algo = self.get_settings("OIDC_USERNAME_ALGO", None)

        if username_algo:
            if isinstance(username_algo, str):
                username_algo = import_string(username_algo)
            if len(inspect.getfullargspec(username_algo).args) == 1:
                # this is for backwards compatibility only
                return username_algo(claims.get("email"))
            else:
                # also pass the claims to the custom user name algo
                return username_algo(claims.get("email"), claims)

        return default_username_algo(claims.get("email"), claims)

    def update_user(self, user, claims):
        """Update existing user with new claims, if necessary save, and return user"""
        return user

    def _verify_jws(self, payload, key):
        """Verify the given JWS payload with the given key and return the payload"""
        jws = JWS.from_compact(payload)

        try:
            alg = jws.signature.combined.alg.name
        except KeyError:
            msg = "No alg value found in header"
            raise SuspiciousOperation(msg)

        if alg != self.OIDC_RP_SIGN_ALGO:
            msg = (
                "The provider algorithm {!r} does not match the client's "
                "OIDC_RP_SIGN_ALGO.".format(alg)
            )
            raise SuspiciousOperation(msg)

        if isinstance(key, str):
            # Use smart_bytes here since the key string comes from settings.
            jwk = JWK.load(smart_bytes(key))
        else:
            # The key is a json returned from the IDP JWKS endpoint.
            jwk = JWK.from_json(key)

        if not jws.verify(jwk):
            msg = "JWS token verification failed."
            raise SuspiciousOperation(msg)

        return jws.payload

    def retrieve_matching_jwk(self, token):
        """Get the signing key by exploring the JWKS endpoint of the OP."""
        response_jwks = requests.get(
            self.OIDC_OP_JWKS_ENDPOINT,
            verify=self.get_settings("OIDC_VERIFY_SSL", True),
            timeout=self.get_settings("OIDC_TIMEOUT", None),
            proxies=self.get_settings("OIDC_PROXY", None),
        )
        response_jwks.raise_for_status()
        jwks = response_jwks.json()

        # Compute the current header from the given token to find a match
        jws = JWS.from_compact(token)
        json_header = jws.signature.protected
        header = Header.json_loads(json_header)

        key = None
        for jwk in jwks["keys"]:
            if import_from_settings("OIDC_VERIFY_KID", True) and jwk[
                "kid"
            ] != smart_str(header.kid):
                continue
            if "alg" in jwk and jwk["alg"] != smart_str(header.alg):
                continue
            key = jwk
        if key is None:
            raise SuspiciousOperation("Could not find a valid JWKS.")
        return key

    def get_payload_data(self, token, key):
        """Helper method to get the payload of the JWT token."""
        if self.get_settings("OIDC_ALLOW_UNSECURED_JWT", False):
            header, payload_data, signature = token.split(b".")
            header = json.loads(smart_str(b64decode(header)))

            # If config allows unsecured JWTs check the header and return the decoded payload
            if "alg" in header and header["alg"] == "none":
                return b64decode(payload_data)

        # By default fallback to verify JWT signatures
        return self._verify_jws(token, key)

    def verify_token(self, token, **kwargs):
        """Validate the token signature."""
        nonce = kwargs.get("nonce")

        token = force_bytes(token)
        if self.OIDC_RP_SIGN_ALGO.startswith("RS") or self.OIDC_RP_SIGN_ALGO.startswith(
            "ES"
        ):
            if self.OIDC_RP_IDP_SIGN_KEY is not None:
                key = self.OIDC_RP_IDP_SIGN_KEY
            else:
                key = self.retrieve_matching_jwk(token)
        else:
            key = self.OIDC_RP_CLIENT_SECRET

        payload_data = self.get_payload_data(token, key)

        # The 'token' will always be a byte string since it's
        # the result of base64.urlsafe_b64decode().
        # The payload is always the result of base64.urlsafe_b64decode().
        # In Python 3 and 2, that's always a byte string.
        # In Python3.6, the json.loads() function can accept a byte string
        # as it will automagically decode it to a unicode string before
        # deserializing https://bugs.python.org/issue17909
        payload = json.loads(payload_data.decode("utf-8"))
        token_nonce = payload.get("nonce")

        if self.get_settings("OIDC_USE_NONCE", True) and nonce != token_nonce:
            msg = "JWT Nonce verification failed."
            raise SuspiciousOperation(msg)
        return payload

    def get_token(self, payload):
        """Return token object as a dictionary."""

        auth = None
        if self.get_settings("OIDC_TOKEN_USE_BASIC_AUTH", False):
            # When Basic auth is defined, create the Auth Header and remove secret from payload.
            user = payload.get("client_id")
            pw = payload.get("client_secret")

            auth = HTTPBasicAuth(user, pw)
            del payload["client_secret"]

        response = requests.post(
            self.OIDC_OP_TOKEN_ENDPOINT,
            data=payload,
            auth=auth,
            verify=self.get_settings("OIDC_VERIFY_SSL", True),
            timeout=self.get_settings("OIDC_TIMEOUT", None),
            proxies=self.get_settings("OIDC_PROXY", None),
        )
        self.raise_token_response_error(response)
        return response.json()

    def raise_token_response_error(self, response):
        """Raises :class:`HTTPError`, if one occurred.
        as per: https://datatracker.ietf.org/doc/html/rfc6749#section-5.2
        """
        # if there wasn't an error all is good
        if response.status_code == 200:
            return
        # otherwise something is up...
        http_error_msg = (
            f"Get Token Error (url: {response.url}, "
            f"status: {response.status_code}, "
            f"body: {response.text})"
        )
        raise HTTPError(http_error_msg, response=response)

    def get_userinfo(self, access_token, id_token, payload):
        """Return user details dictionary. The id_token and payload are not used in
        the default implementation, but may be used when overriding this method"""

        user_response = requests.get(
            self.OIDC_OP_USER_ENDPOINT,
            headers={"Authorization": "Bearer {0}".format(access_token)},
            verify=self.get_settings("OIDC_VERIFY_SSL", True),
            timeout=self.get_settings("OIDC_TIMEOUT", None),
            proxies=self.get_settings("OIDC_PROXY", None),
        )
        user_response.raise_for_status()
        return user_response.json()

    def authenticate(self, request, **kwargs):
        """Authenticates a user based on the OIDC code flow."""

        self.request = request
        if not self.request:
            return None

        state = self.request.GET.get("state")
        code = self.request.GET.get("code")
        nonce = kwargs.pop("nonce", None)
        code_verifier = kwargs.pop("code_verifier", None)

        if not code or not state:
            return None

        reverse_url = self.get_settings(
            "OIDC_AUTHENTICATION_CALLBACK_URL", "oidc_authentication_callback"
        )

        token_payload = {
            "client_id": self.OIDC_RP_CLIENT_ID,
            "client_secret": self.OIDC_RP_CLIENT_SECRET,
            "grant_type": "authorization_code",
            "code": code,
            "redirect_uri": absolutify(self.request, reverse(reverse_url)),
        }

        # Send code_verifier with token request if using PKCE
        if code_verifier is not None:
            token_payload.update({"code_verifier": code_verifier})

        # Get the token
        token_info = self.get_token(token_payload)
        id_token = token_info.get("id_token")
        access_token = token_info.get("access_token")

        # Validate the token
        payload = self.verify_token(id_token, nonce=nonce)

        if payload:
            self.store_tokens(access_token, id_token)
            try:
                return self.get_or_create_user(access_token, id_token, payload)
            except SuspiciousOperation as exc:
                LOGGER.warning("failed to get or create user: %s", exc)
                return None

        return None

    def store_tokens(self, access_token, id_token):
        """Store OIDC tokens."""
        session = self.request.session

        if self.get_settings("OIDC_STORE_ACCESS_TOKEN", False):
            session["oidc_access_token"] = access_token

        if self.get_settings("OIDC_STORE_ID_TOKEN", False):
            session["oidc_id_token"] = id_token

    def get_or_create_user(self, access_token, id_token, payload):
        """Returns a User instance if 1 user is found. Creates a user if not found
        and configured to do so. Returns nothing if multiple users are matched."""

        user_info = self.get_userinfo(access_token, id_token, payload)

        claims_verified = self.verify_claims(user_info)
        if not claims_verified:
            msg = "Claims verification failed"
            raise SuspiciousOperation(msg)

        # email based filtering
        users = self.filter_users_by_claims(user_info)

        if len(users) == 1:
            return self.update_user(users[0], user_info)
        elif len(users) > 1:
            # In the rare case that two user accounts have the same email address,
            # bail. Randomly selecting one seems really wrong.
            msg = "Multiple users returned"
            raise SuspiciousOperation(msg)
        elif self.get_settings("OIDC_CREATE_USER", True):
            user = self.create_user(user_info)
            return user
        else:
            LOGGER.debug(
                "Login failed: No user with %s found, and " "OIDC_CREATE_USER is False",
                self.describe_user_by_claims(user_info),
            )
            return None

    def get_user(self, user_id):
        """Return a user based on the id."""

        try:
            return self.UserModel.objects.get(pk=user_id)
        except self.UserModel.DoesNotExist:
            return None



import json
import re
import time
from urllib.parse import parse_qs

from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.contrib.auth.signals import user_logged_out
from django.core.cache import cache
from django.dispatch import receiver
from django.http import HttpResponse
from django.test import Client, RequestFactory, TestCase, override_settings
from django.test.client import ClientHandler
from django.urls import path
from unittest.mock import MagicMock, patch

from mozilla_django_oidc.middleware import SessionRefresh
from mozilla_django_oidc.urls import urlpatterns as orig_urlpatterns

User = get_user_model()


@override_settings(OIDC_OP_AUTHORIZATION_ENDPOINT="http://example.com/authorize")
@override_settings(OIDC_RP_CLIENT_ID="foo")
@override_settings(OIDC_RENEW_ID_TOKEN_EXPIRY_SECONDS=120)
@patch("mozilla_django_oidc.middleware.get_random_string")
class SessionRefreshTokenMiddlewareTestCase(TestCase):
    def setUp(self):
        self.factory = RequestFactory()
        self.middleware = SessionRefresh(MagicMock)
        self.user = User.objects.create_user("example_username")

    def test_anonymous(self, mock_middleware_random):
        request = self.factory.get("/foo")
        request.session = {}
        request.user = AnonymousUser()
        response = self.middleware.process_request(request)
        self.assertTrue(not response)

    def test_is_oidc_path(self, mock_middleware_random):
        request = self.factory.get("/oidc/callback/")
        request.user = AnonymousUser()
        request.session = {}
        response = self.middleware.process_request(request)
        self.assertTrue(not response)

    def test_is_POST(self, mock_middleware_random):
        request = self.factory.post("/foo")
        request.user = AnonymousUser()
        request.session = {}
        response = self.middleware.process_request(request)
        self.assertTrue(not response)

    def test_is_ajax(self, mock_middleware_random):
        mock_middleware_random.return_value = "examplestring"

        request = self.factory.get("/foo", HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        request.session = {}
        request.user = self.user

        response = self.middleware.process_request(request)
        self.assertEqual(response.status_code, 403)
        # The URL to go to is available both as a header and as a key
        # in the JSON response.
        self.assertTrue(response["refresh_url"])
        url, qs = response["refresh_url"].split("?")
        self.assertEqual(url, "http://example.com/authorize")
        expected_query = {
            "response_type": ["code"],
            "redirect_uri": ["http://testserver/callback/"],
            "client_id": ["foo"],
            "nonce": ["examplestring"],
            "prompt": ["none"],
            "scope": ["openid email"],
            "state": ["examplestring"],
        }
        self.assertEqual(expected_query, parse_qs(qs))
        json_payload = json.loads(response.content.decode("utf-8"))
        self.assertEqual(json_payload["refresh_url"], response["refresh_url"])

    @override_settings(OIDC_USE_PKCE=True)
    def test_is_ajax_with_pkce(self, mock_middleware_random):
        mock_middleware_random.return_value = "examplestring"

        request = self.factory.get("/foo", HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        request.session = {}
        request.user = self.user

        response = self.middleware.process_request(request)
        self.assertEqual(response.status_code, 403)
        # The URL to go to is available both as a header and as a key
        # in the JSON response.
        self.assertTrue(response["refresh_url"])
        url, qs = response["refresh_url"].split("?")
        self.assertEqual(url, "http://example.com/authorize")
        expected_query = {
            "response_type": ["code"],
            "redirect_uri": ["http://testserver/callback/"],
            "client_id": ["foo"],
            "nonce": ["examplestring"],
            "prompt": ["none"],
            "scope": ["openid email"],
            "state": ["examplestring"],
            "code_challenge_method": ["S256"],
            "code_challenge": ["m8yog7rVNdOd7hYIoUg6yl5mk_IYauWdSIBUjoPJHB0"],
        }
        self.assertEqual(expected_query, parse_qs(qs))
        json_payload = json.loads(response.content.decode("utf-8"))
        self.assertEqual(json_payload["refresh_url"], response["refresh_url"])

    def test_no_oidc_token_expiration_forces_renewal(self, mock_middleware_random):
        mock_middleware_random.return_value = "examplestring"

        request = self.factory.get("/foo")
        request.user = self.user
        request.session = {}

        response = self.middleware.process_request(request)

        self.assertEqual(response.status_code, 302)
        url, qs = response.url.split("?")
        self.assertEqual(url, "http://example.com/authorize")
        expected_query = {
            "response_type": ["code"],
            "redirect_uri": ["http://testserver/callback/"],
            "client_id": ["foo"],
            "nonce": ["examplestring"],
            "prompt": ["none"],
            "scope": ["openid email"],
            "state": ["examplestring"],
        }
        self.assertEqual(expected_query, parse_qs(qs))

    @override_settings(OIDC_USE_PKCE=True)
    def test_no_oidc_token_expiration_forces_renewal_with_pkce(
        self, mock_middleware_random
    ):
        mock_middleware_random.return_value = "examplestring"

        request = self.factory.get("/foo")
        request.user = self.user
        request.session = {}

        response = self.middleware.process_request(request)

        self.assertEqual(response.status_code, 302)
        url, qs = response.url.split("?")
        self.assertEqual(url, "http://example.com/authorize")
        expected_query = {
            "response_type": ["code"],
            "redirect_uri": ["http://testserver/callback/"],
            "client_id": ["foo"],
            "nonce": ["examplestring"],
            "prompt": ["none"],
            "scope": ["openid email"],
            "state": ["examplestring"],
            "code_challenge_method": ["S256"],
            "code_challenge": ["m8yog7rVNdOd7hYIoUg6yl5mk_IYauWdSIBUjoPJHB0"],
        }
        self.assertEqual(expected_query, parse_qs(qs))

    def test_expired_token_forces_renewal(self, mock_middleware_random):
        mock_middleware_random.return_value = "examplestring"

        request = self.factory.get("/foo")
        request.user = self.user
        request.session = {"oidc_id_token_expiration": time.time() - 10}

        response = self.middleware.process_request(request)

        self.assertEqual(response.status_code, 302)
        url, qs = response.url.split("?")
        self.assertEqual(url, "http://example.com/authorize")
        expected_query = {
            "response_type": ["code"],
            "redirect_uri": ["http://testserver/callback/"],
            "client_id": ["foo"],
            "nonce": ["examplestring"],
            "prompt": ["none"],
            "scope": ["openid email"],
            "state": ["examplestring"],
        }
        self.assertEqual(expected_query, parse_qs(qs))

    @override_settings(OIDC_USE_PKCE=True)
    def test_expired_token_forces_renewal_with_pkce(self, mock_middleware_random):
        mock_middleware_random.return_value = "examplestring"

        request = self.factory.get("/foo")
        request.user = self.user
        request.session = {"oidc_id_token_expiration": time.time() - 10}

        response = self.middleware.process_request(request)

        self.assertEqual(response.status_code, 302)
        url, qs = response.url.split("?")
        self.assertEqual(url, "http://example.com/authorize")
        expected_query = {
            "response_type": ["code"],
            "redirect_uri": ["http://testserver/callback/"],
            "client_id": ["foo"],
            "nonce": ["examplestring"],
            "prompt": ["none"],
            "scope": ["openid email"],
            "state": ["examplestring"],
            "code_challenge_method": ["S256"],
            "code_challenge": ["m8yog7rVNdOd7hYIoUg6yl5mk_IYauWdSIBUjoPJHB0"],
        }
        self.assertEqual(expected_query, parse_qs(qs))


# This adds a "home page" we can test against.
def fakeview(req):
    return HttpResponse("Win!")


urlpatterns = list(orig_urlpatterns) + [
    path("mdo_fake_view/", fakeview, name="mdo_fake_view")
]


def override_middleware(fun):
    classes = [
        "django.contrib.sessions.middleware.SessionMiddleware",
        "mozilla_django_oidc.middleware.SessionRefresh",
    ]
    return override_settings(MIDDLEWARE=classes)(fun)


class UserifiedClientHandler(ClientHandler):
    """Enhances ClientHandler to "work" with users properly"""

    def __init__(self, *args, **kwargs):
        self.user = kwargs.pop("user")
        super(UserifiedClientHandler, self).__init__(*args, **kwargs)

    def get_response(self, req):
        req.user = self.user
        return super(UserifiedClientHandler, self).get_response(req)


class ClientWithUser(Client):
    """Enhances Client to "work" with users properly"""

    def __init__(self, enforce_csrf_checks=False, **defaults):
        # Start off with the AnonymousUser
        self.user = AnonymousUser()
        # Get this because we need to create a new UserifiedClientHandler later
        self.enforce_csrf_checks = enforce_csrf_checks
        super(ClientWithUser, self).__init__(**defaults)
        # Stomp on the ClientHandler with one that correctly makes request.user
        # the AnonymousUser
        self.handler = UserifiedClientHandler(enforce_csrf_checks, user=self.user)

    def login(self, **credentials):
        from django.contrib.auth import authenticate

        # Try to authenticate and throw an exception if that fails; also, this gets
        # the user instance that was authenticated with
        user = authenticate(**credentials)
        if not user:
            # Client lets you fail authentication without providing any helpful
            # messages; we throw an exception because silent failure is
            # unhelpful
            raise Exception("Unable to authenticate with %r" % credentials)

        ret = super(ClientWithUser, self).login(**credentials)
        if not ret:
            raise Exception("Login failed")

        # Stash the user object it used and rebuild the UserifiedClientHandler
        self.user = user
        self.handler = UserifiedClientHandler(self.enforce_csrf_checks, user=self.user)
        return ret


@override_settings(OIDC_RP_CLIENT_ID="foo")
@override_settings(OIDC_RENEW_ID_TOKEN_EXPIRY_SECONDS=120)
@override_settings(OIDC_OP_AUTHORIZATION_ENDPOINT="http://example.com/authorize")
@override_settings(ROOT_URLCONF="tests.test_middleware")
@override_middleware
class MiddlewareTestCase(TestCase):
    """These tests test the middleware as part of the request/response cycle"""

    def setUp(self):
        self.factory = RequestFactory()
        self.user = User.objects.create_user(
            username="example_username", password="password"
        )
        cache.clear()

    @override_settings(OIDC_EXEMPT_URLS=["mdo_fake_view"])
    def test_get_exempt_urls_setting_view_name(self):
        middleware = SessionRefresh(MagicMock())
        self.assertEqual(
            sorted(list(middleware.exempt_urls)),
            ["/authenticate/", "/callback/", "/logout/", "/mdo_fake_view/"],
        )

    @override_settings(OIDC_EXEMPT_URLS=["/foo/"])
    def test_get_exempt_urls_setting_url_path(self):
        middleware = SessionRefresh(MagicMock())
        self.assertEqual(
            sorted(list(middleware.exempt_urls)),
            ["/authenticate/", "/callback/", "/foo/", "/logout/"],
        )

    def test_is_refreshable_url(self):
        request = self.factory.get("/mdo_fake_view/")
        request.user = self.user
        request.session = dict()
        middleware = SessionRefresh(MagicMock())
        assert middleware.is_refreshable_url(request)

    @override_settings(OIDC_EXEMPT_URLS=["mdo_fake_view"])
    def test_is_not_refreshable_url_exempt_view_name(self):
        request = self.factory.get("/mdo_fake_view/")
        request.user = self.user
        request.session = dict()
        middleware = SessionRefresh(MagicMock())
        assert not middleware.is_refreshable_url(request)

    @override_settings(OIDC_EXEMPT_URLS=["/mdo_fake_view/"])
    def test_is_not_refreshable_url_exempt_path(self):
        request = self.factory.get("/mdo_fake_view/")
        request.user = self.user
        request.session = dict()
        middleware = SessionRefresh(MagicMock())
        assert not middleware.is_refreshable_url(request)

    @override_settings(OIDC_EXEMPT_URLS=[re.compile(r"^/mdo_.*_view/$")])
    def test_is_not_refreshable_url_exempt_pattern(self):
        request = self.factory.get("/mdo_fake_view/")
        request.user = self.user
        request.session = dict()
        middleware = SessionRefresh(MagicMock())
        assert not middleware.is_refreshable_url(request)

    def test_anonymous(self):
        client = ClientWithUser()
        resp = client.get("/mdo_fake_view/")
        self.assertEqual(resp.status_code, 200)

    @override_settings(OIDC_OP_AUTHORIZATION_ENDPOINT="http://example.com/authorize")
    @override_settings(OIDC_RP_CLIENT_ID="foo")
    @override_settings(OIDC_RENEW_ID_TOKEN_EXPIRY_SECONDS=120)
    def test_authenticated_user(self):
        client = ClientWithUser()
        client.login(username=self.user.username, password="password")

        # Set the expiration to some time in the future so this user is valid
        session = client.session
        session["oidc_id_token_expiration"] = time.time() + 100
        session.save()

        resp = client.get("/mdo_fake_view/")
        self.assertEqual(resp.status_code, 200)

    @override_settings(OIDC_OP_AUTHORIZATION_ENDPOINT="http://example.com/authorize")
    @override_settings(OIDC_RP_CLIENT_ID="foo")
    @override_settings(OIDC_RENEW_ID_TOKEN_EXPIRY_SECONDS=120)
    @patch("mozilla_django_oidc.middleware.get_random_string")
    def test_expired_token_redirects_to_sso(self, mock_middleware_random):
        mock_middleware_random.return_value = "examplestring"

        client = ClientWithUser()
        client.login(username=self.user.username, password="password")

        # Set expiration to some time in the past
        session = client.session
        session["oidc_id_token_expiration"] = time.time() - 100
        session[
            "_auth_user_backend"
        ] = "mozilla_django_oidc.auth.OIDCAuthenticationBackend"
        session.save()

        resp = client.get("/mdo_fake_view/")
        self.assertEqual(resp.status_code, 302)

        url, qs = resp.url.split("?")
        self.assertEqual(url, "http://example.com/authorize")
        expected_query = {
            "response_type": ["code"],
            "redirect_uri": ["http://testserver/callback/"],
            "client_id": ["foo"],
            "nonce": ["examplestring"],
            "prompt": ["none"],
            "scope": ["openid email"],
            "state": ["examplestring"],
        }
        self.assertEqual(expected_query, parse_qs(qs))

    @override_settings(OIDC_OP_AUTHORIZATION_ENDPOINT="http://example.com/authorize")
    @override_settings(OIDC_RP_CLIENT_ID="foo")
    @override_settings(OIDC_RENEW_ID_TOKEN_EXPIRY_SECONDS=120)
    @override_settings(OIDC_USE_PKCE=True)
    @patch("mozilla_django_oidc.middleware.get_random_string")
    def test_expired_token_redirects_to_sso_with_pkce(self, mock_middleware_random):
        mock_middleware_random.return_value = "examplestring"

        client = ClientWithUser()
        client.login(username=self.user.username, password="password")

        # Set expiration to some time in the past
        session = client.session
        session["oidc_id_token_expiration"] = time.time() - 100
        session[
            "_auth_user_backend"
        ] = "mozilla_django_oidc.auth.OIDCAuthenticationBackend"
        session.save()

        resp = client.get("/mdo_fake_view/")
        self.assertEqual(resp.status_code, 302)

        url, qs = resp.url.split("?")
        self.assertEqual(url, "http://example.com/authorize")
        expected_query = {
            "response_type": ["code"],
            "redirect_uri": ["http://testserver/callback/"],
            "client_id": ["foo"],
            "nonce": ["examplestring"],
            "prompt": ["none"],
            "scope": ["openid email"],
            "state": ["examplestring"],
            "code_challenge_method": ["S256"],
            "code_challenge": ["m8yog7rVNdOd7hYIoUg6yl5mk_IYauWdSIBUjoPJHB0"],
        }
        self.assertEqual(expected_query, parse_qs(qs))

    @override_settings(OIDC_OP_AUTHORIZATION_ENDPOINT="http://example.com/authorize")
    @override_settings(OIDC_RP_CLIENT_ID="foo")
    @override_settings(OIDC_RENEW_ID_TOKEN_EXPIRY_SECONDS=120)
    @patch("mozilla_django_oidc.middleware.get_random_string")
    def test_refresh_fails_for_already_signed_in_user(self, mock_random_string):
        mock_random_string.return_value = "examplestring"

        # Mutable to log which users get logged out.
        logged_out_users = []

        # Register a signal on 'user_logged_out' so we can
        # update 'logged_out_users'.
        @receiver(user_logged_out)
        def logged_out(sender, user=None, **kwargs):
            logged_out_users.append(user)

        client = ClientWithUser()
        # First confirm that the home page is a public page.
        resp = client.get("/")
        # At least security doesn't kick you out.
        self.assertEqual(resp.status_code, 404)
        # Also check that this page doesn't force you to redirect
        # to authenticate.
        resp = client.get("/mdo_fake_view/")
        self.assertEqual(resp.status_code, 200)
        client.login(username=self.user.username, password="password")

        # Set expiration to some time in the past
        session = client.session
        session["oidc_id_token_expiration"] = time.time() - 100
        session[
            "_auth_user_backend"
        ] = "mozilla_django_oidc.auth.OIDCAuthenticationBackend"
        session.save()

        # Confirm that now you're forced to authenticate again.
        resp = client.get("/mdo_fake_view/")
        self.assertEqual(resp.status_code, 302)
        self.assertTrue(
            "http://example.com/authorize" in resp.url and "prompt=none" in resp.url
        )
        # Now suppose the user goes there and something goes wrong.
        # For example, the user might have become "blocked" or the 2FA
        # verficiation has expired and needs to be done again.
        resp = client.get(
            "/callback/",
            {
                "error": "login_required",
                "error_description": "Multifactor authentication required",
            },
        )
        self.assertEqual(resp.status_code, 302)
        self.assertEqual(resp.url, "/")

        # Since the user in 'client' doesn't change, we have to use other
        # queues to assert that the user got logged out properly.

        # The session gets flushed when you get signed out.
        # This is the only decent way to know the user lost all
        # request.session and
        self.assertTrue(not client.session.items())

        # The signal we registered should have fired for this user.
        self.assertEqual(client.user, logged_out_users[0])



"""
Classes/functions for integrating with Django REST Framework.

http://www.django-rest-framework.org/api-guide/authentication/#custom-authentication
"""

import logging

from django.contrib.auth import get_backends
from django.core.exceptions import ImproperlyConfigured, SuspiciousOperation
from django.utils.module_loading import import_string
from rest_framework import authentication, exceptions
from requests.exceptions import HTTPError

from mozilla_django_oidc.auth import OIDCAuthenticationBackend
from mozilla_django_oidc.utils import (
    import_from_settings,
    parse_www_authenticate_header,
)

LOGGER = logging.getLogger(__name__)


def get_oidc_backend():
    """
    Get the Django auth backend that uses OIDC.
    """

    # allow the user to force which back backend to use. this is mostly
    # convenient if you want to use OIDC with DRF but don't want to configure
    # OIDC for the "normal" Django auth.
    backend_setting = import_from_settings("OIDC_DRF_AUTH_BACKEND", None)
    if backend_setting:
        backend = import_string(backend_setting)()
        if not isinstance(backend, OIDCAuthenticationBackend):
            msg = (
                "Class configured in OIDC_DRF_AUTH_BACKEND "
                "does not extend OIDCAuthenticationBackend!"
            )
            raise ImproperlyConfigured(msg)
        return backend

    # if the backend setting is not set, look through the list of configured
    # backends for one that is an OIDCAuthenticationBackend.
    backends = [b for b in get_backends() if isinstance(b, OIDCAuthenticationBackend)]

    if not backends:
        msg = (
            "No backends extending OIDCAuthenticationBackend found - "
            "add one to AUTHENTICATION_BACKENDS or set OIDC_DRF_AUTH_BACKEND!"
        )
        raise ImproperlyConfigured(msg)
    if len(backends) > 1:
        raise ImproperlyConfigured("More than one OIDCAuthenticationBackend found!")
    return backends[0]


class OIDCAuthentication(authentication.BaseAuthentication):
    """
    Provide OpenID authentication for DRF.
    """

    # used by the authenticate_header method.
    www_authenticate_realm = "api"

    def __init__(self, backend=None):
        self.backend = backend or get_oidc_backend()

    def authenticate(self, request):
        """
        Authenticate the request and return a tuple of (user, token) or None
        if there was no authentication attempt.
        """
        access_token = self.get_access_token(request)

        if not access_token:
            return None

        try:
            user = self.backend.get_or_create_user(access_token, None, None)
        except HTTPError as exc:
            resp = exc.response

            # if the oidc provider returns 401, it means the token is invalid.
            # in that case, we want to return the upstream error message (which
            # we can get from the www-authentication header) in the response.
            if resp.status_code == 401 and "www-authenticate" in resp.headers:
                data = parse_www_authenticate_header(resp.headers["www-authenticate"])
                raise exceptions.AuthenticationFailed(
                    data.get(
                        "error_description", "no error description in www-authenticate"
                    )
                )

            # for all other http errors, just re-raise the exception.
            raise
        except SuspiciousOperation as exc:
            LOGGER.info("Login failed: %s", exc)
            raise exceptions.AuthenticationFailed("Login failed")

        if not user:
            msg = "Login failed: No user found for the given access token."
            raise exceptions.AuthenticationFailed(msg)

        return user, access_token

    def get_access_token(self, request):
        """
        Get the access token based on a request.

        Returns None if no authentication details were provided. Raises
        AuthenticationFailed if the token is incorrect.
        """
        header = authentication.get_authorization_header(request)
        if not header:
            return None
        header = header.decode(authentication.HTTP_HEADER_ENCODING)

        auth = header.split()

        if auth[0].lower() != "bearer":
            return None

        if len(auth) == 1:
            msg = 'Invalid "bearer" header: No credentials provided.'
            raise exceptions.AuthenticationFailed(msg)
        elif len(auth) > 2:
            msg = (
                'Invalid "bearer" header: Credentials string should not contain spaces.'
            )
            raise exceptions.AuthenticationFailed(msg)

        return auth[1]

    def authenticate_header(self, request):
        """
        If this method returns None, a generic HTTP 403 forbidden response is
        returned by DRF when authentication fails.

        By making the method return a string, a 401 is returned instead. The
        return value will be used as the WWW-Authenticate header.
        """
        return 'Bearer realm="%s"' % self.www_authenticate_realm



import time
from urllib.parse import urlencode

from django.contrib import auth
from django.core.exceptions import SuspiciousOperation
from django.http import HttpResponseNotAllowed, HttpResponseRedirect
from django.shortcuts import resolve_url
from django.urls import reverse
from django.utils.crypto import get_random_string
from django.utils.http import url_has_allowed_host_and_scheme
from django.utils.module_loading import import_string
from django.views.generic import View

from mozilla_django_oidc.utils import (
    absolutify,
    add_state_and_verifier_and_nonce_to_session,
    generate_code_challenge,
    import_from_settings,
)


class OIDCAuthenticationCallbackView(View):
    """OIDC client authentication callback HTTP endpoint"""

    http_method_names = ["get"]

    @staticmethod
    def get_settings(attr, *args):
        return import_from_settings(attr, *args)

    @property
    def failure_url(self):
        return self.get_settings("LOGIN_REDIRECT_URL_FAILURE", "/")

    @property
    def success_url(self):
        # Pull the next url from the session or settings--we don't need to
        # sanitize here because it should already have been sanitized.
        next_url = self.request.session.get("oidc_login_next", None)
        return next_url or resolve_url(self.get_settings("LOGIN_REDIRECT_URL", "/"))

    def login_failure(self):
        return HttpResponseRedirect(self.failure_url)

    def login_success(self):
        # If the user hasn't changed (because this is a session refresh instead of a
        # normal login), don't call login. This prevents invaliding the user's current CSRF token
        request_user = getattr(self.request, "user", None)
        if (
            not request_user
            or not request_user.is_authenticated
            or request_user != self.user
        ):
            auth.login(self.request, self.user)

        # Figure out when this id_token will expire. This is ignored unless you're
        # using the SessionRefresh middleware.
        expiration_interval = self.get_settings(
            "OIDC_RENEW_ID_TOKEN_EXPIRY_SECONDS", 60 * 15
        )
        self.request.session["oidc_id_token_expiration"] = (
            time.time() + expiration_interval
        )

        return HttpResponseRedirect(self.success_url)

    def get(self, request):
        """Callback handler for OIDC authorization code flow"""

        if request.GET.get("error"):
            # Ouch! Something important failed.

            # Delete the state entry also for failed authentication attempts
            # to prevent replay attacks.
            if (
                "state" in request.GET
                and "oidc_states" in request.session
                and request.GET["state"] in request.session["oidc_states"]
            ):
                del request.session["oidc_states"][request.GET["state"]]
                request.session.save()

            # Make sure the user doesn't get to continue to be logged in
            # otherwise the refresh middleware will force the user to
            # redirect to authorize again if the session refresh has
            # expired.
            if request.user.is_authenticated:
                auth.logout(request)
            assert not request.user.is_authenticated
        elif "code" in request.GET and "state" in request.GET:
            # Check instead of "oidc_state" check if the "oidc_states" session key exists!
            if "oidc_states" not in request.session:
                return self.login_failure()

            # State, Nonce and PKCE Code Verifier are stored in the session "oidc_states"
            # dictionary.
            # State is the key, the value is a dictionary with the Nonce in the "nonce" field, and
            # Code Verifier or None in the "code_verifier" field.
            state = request.GET.get("state")
            if state not in request.session["oidc_states"]:
                msg = "OIDC callback state not found in session `oidc_states`!"
                raise SuspiciousOperation(msg)

            # Get the nonce and optional code verifier from the dictionary for further processing
            # and delete the entry to prevent replay attacks.
            code_verifier = request.session["oidc_states"][state].get("code_verifier")
            nonce = request.session["oidc_states"][state]["nonce"]
            del request.session["oidc_states"][state]

            # Authenticating is slow, so save the updated oidc_states.
            request.session.save()
            # Reset the session. This forces the session to get reloaded from the database after
            # fetching the token from the OpenID connect provider.
            # Without this step we would overwrite items that are being added/removed from the
            # session in parallel browser tabs.
            request.session = request.session.__class__(request.session.session_key)

            kwargs = {
                "request": request,
                "nonce": nonce,
                "code_verifier": code_verifier,
            }

            self.user = auth.authenticate(**kwargs)

            if self.user and self.user.is_active:
                return self.login_success()
        return self.login_failure()


def get_next_url(request, redirect_field_name):
    """Retrieves next url from request

    Note: This verifies that the url is safe before returning it. If the url
    is not safe, this returns None.

    :arg HttpRequest request: the http request
    :arg str redirect_field_name: the name of the field holding the next url

    :returns: safe url or None

    """
    next_url = request.GET.get(redirect_field_name)
    if next_url:
        kwargs = {
            "url": next_url,
            "require_https": import_from_settings(
                "OIDC_REDIRECT_REQUIRE_HTTPS", request.is_secure()
            ),
        }

        hosts = list(import_from_settings("OIDC_REDIRECT_ALLOWED_HOSTS", []))
        hosts.append(request.get_host())
        kwargs["allowed_hosts"] = hosts

        is_safe = url_has_allowed_host_and_scheme(**kwargs)
        if is_safe:
            return next_url
    return None


class OIDCAuthenticationRequestView(View):
    """OIDC client authentication HTTP endpoint"""

    http_method_names = ["get"]

    def __init__(self, *args, **kwargs):
        super(OIDCAuthenticationRequestView, self).__init__(*args, **kwargs)

        self.OIDC_OP_AUTH_ENDPOINT = self.get_settings("OIDC_OP_AUTHORIZATION_ENDPOINT")
        self.OIDC_RP_CLIENT_ID = self.get_settings("OIDC_RP_CLIENT_ID")

    @staticmethod
    def get_settings(attr, *args):
        return import_from_settings(attr, *args)

    def get(self, request):
        """OIDC client authentication initialization HTTP endpoint"""
        state = get_random_string(self.get_settings("OIDC_STATE_SIZE", 32))
        redirect_field_name = self.get_settings("OIDC_REDIRECT_FIELD_NAME", "next")
        reverse_url = self.get_settings(
            "OIDC_AUTHENTICATION_CALLBACK_URL", "oidc_authentication_callback"
        )

        params = {
            "response_type": "code",
            "scope": self.get_settings("OIDC_RP_SCOPES", "openid email"),
            "client_id": self.OIDC_RP_CLIENT_ID,
            "redirect_uri": absolutify(request, reverse(reverse_url)),
            "state": state,
        }

        params.update(self.get_extra_params(request))

        if self.get_settings("OIDC_USE_NONCE", True):
            nonce = get_random_string(self.get_settings("OIDC_NONCE_SIZE", 32))
            params.update({"nonce": nonce})

        if self.get_settings("OIDC_USE_PKCE", False):
            code_verifier_length = self.get_settings("OIDC_PKCE_CODE_VERIFIER_SIZE", 64)
            # Check that code_verifier_length is between the min and max length
            # defined in https://datatracker.ietf.org/doc/html/rfc7636#section-4.1
            if not (43 <= code_verifier_length <= 128):
                raise ValueError("code_verifier_length must be between 43 and 128")

            # Generate code_verifier and code_challenge pair
            code_verifier = get_random_string(code_verifier_length)
            code_challenge_method = self.get_settings(
                "OIDC_PKCE_CODE_CHALLENGE_METHOD", "S256"
            )
            code_challenge = generate_code_challenge(
                code_verifier, code_challenge_method
            )

            # Append code_challenge to authentication request parameters
            params.update(
                {
                    "code_challenge": code_challenge,
                    "code_challenge_method": code_challenge_method,
                }
            )

        else:
            code_verifier = None

        add_state_and_verifier_and_nonce_to_session(
            request, state, params, code_verifier
        )

        request.session["oidc_login_next"] = get_next_url(request, redirect_field_name)

        query = urlencode(params)
        redirect_url = "{url}?{query}".format(
            url=self.OIDC_OP_AUTH_ENDPOINT, query=query
        )
        return HttpResponseRedirect(redirect_url)

    def get_extra_params(self, request):
        return self.get_settings("OIDC_AUTH_REQUEST_EXTRA_PARAMS", {})


class OIDCLogoutView(View):
    """Logout helper view"""

    http_method_names = ["get", "post"]

    @staticmethod
    def get_settings(attr, *args):
        return import_from_settings(attr, *args)

    @property
    def redirect_url(self):
        """Return the logout url defined in settings."""
        return self.get_settings("LOGOUT_REDIRECT_URL", "/")

    def post(self, request):
        """Log out the user."""
        logout_url = self.redirect_url

        if request.user.is_authenticated:
            # Check if a method exists to build the URL to log out the user
            # from the OP.
            logout_from_op = self.get_settings("OIDC_OP_LOGOUT_URL_METHOD", "")
            if logout_from_op:
                logout_url = import_string(logout_from_op)(request)

            # Log out the Django user if they were logged in.
            auth.logout(request)

        return HttpResponseRedirect(logout_url)

    def get(self, request):
        """Log out the user."""
        if self.get_settings("ALLOW_LOGOUT_GET_METHOD", False):
            return self.post(request)
        return HttpResponseNotAllowed(["POST"])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

from mozilla_django_oidc import __version__ as VERSION

if sys.argv[-1] == "publish":
    try:
        import wheel

        print("Wheel version: ", wheel.__version__)
    except ImportError:
        print('Wheel library missing. Please run "pip install wheel"')
        sys.exit()
    os.system("python setup.py sdist upload")
    os.system("python setup.py bdist_wheel upload")
    sys.exit()

if sys.argv[-1] == "tag":
    print("Tagging the version on git:")
    os.system("git tag -a %s -m 'version %s'" % (VERSION, VERSION))
    os.system("git push --tags")
    sys.exit()

readme = open("README.rst").read()
history = open("HISTORY.rst").read().replace(".. :changelog:", "")

install_requirements = [
    "Django >= 3.2",
    "josepy",
    "requests",
    "cryptography",
]

setup(
    name="mozilla-django-oidc",
    version=VERSION,
    description="""A lightweight authentication and access management library for integration with OpenID Connect enabled authentication services.""",  # noqa
    long_description=readme + "\n\n" + history,
    author="Tasos Katsoulas, John Giannelos",
    author_email="akatsoulas@mozilla.com, jgiannelos@mozilla.com",
    url="https://github.com/mozilla/mozilla-django-oidc",
    packages=["mozilla_django_oidc"],
    include_package_data=True,
    install_requires=install_requirements,
    license="MPL 2.0",
    zip_safe=False,
    keywords="mozilla-django-oidc",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: Django",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.2",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Intended Audience :: Developers",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: Linux",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
)
